@@ -1,68 +1,109 @@
 """Package with general repository related functions"""
+from git.refs.tag import Tag
 import os
+import stat
 from string import digits
 
+from git.exc import WorkTreeRepositoryUnsupported
+from git.objects import Object
+from git.refs import SymbolicReference
+from git.util import hex_to_bin, bin_to_hex, decygpath
 from gitdb.exc import (
     BadObject,
-    BadName
+    BadName,
 )
-from git.refs import SymbolicReference
-from git.objects import Object
-from gitdb.util import (
-    join,
-    isdir,
-    isfile,
-    dirname,
-    hex_to_bin,
-    bin_to_hex
-)
-from git.compat import xrange
+
+import os.path as osp
+from git.cmd import Git
+
+# Typing ----------------------------------------------------------------------
+
+from typing import AnyStr, Union, Optional, cast, TYPE_CHECKING
+from git.types import PathLike
+if TYPE_CHECKING:
+    from .base import Repo
+    from git.db import GitCmdObjectDB
+    from git.objects import Commit, TagObject, Blob, Tree
+
+# ----------------------------------------------------------------------------
+
+__all__ = ('rev_parse', 'is_git_dir', 'touch', 'find_submodule_git_dir', 'name_to_object', 'short_to_long', 'deref_tag',
+           'to_commit', 'find_worktree_git_dir')
 
 
-__all__ = ('rev_parse', 'is_git_dir', 'touch', 'find_git_dir', 'name_to_object', 'short_to_long', 'deref_tag',
-           'to_commit')
-
-
-def touch(filename):
-    fp = open(filename, "ab")
-    fp.close()
+def touch(filename: str) -> str:
+    with open(filename, "ab"):
+        pass
     return filename
 
 
-def is_git_dir(d):
+def is_git_dir(d: PathLike) -> bool:
     """ This is taken from the git setup.c:is_git_directory
-    function."""
-    if isdir(d) and \
-            isdir(join(d, 'objects')) and \
-            isdir(join(d, 'refs')):
-        headref = join(d, 'HEAD')
-        return isfile(headref) or \
-            (os.path.islink(headref) and
-             os.readlink(headref).startswith('refs'))
+    function.
+
+    @throws WorkTreeRepositoryUnsupported if it sees a worktree directory. It's quite hacky to do that here,
+            but at least clearly indicates that we don't support it.
+            There is the unlikely danger to throw if we see directories which just look like a worktree dir,
+            but are none."""
+    if osp.isdir(d):
+        if (osp.isdir(osp.join(d, 'objects')) or 'GIT_OBJECT_DIRECTORY' in os.environ) \
+           and osp.isdir(osp.join(d, 'refs')):
+            headref = osp.join(d, 'HEAD')
+            return osp.isfile(headref) or \
+                (osp.islink(headref) and
+                 os.readlink(headref).startswith('refs'))
+        elif (osp.isfile(osp.join(d, 'gitdir')) and
+              osp.isfile(osp.join(d, 'commondir')) and
+              osp.isfile(osp.join(d, 'gitfile'))):
+            raise WorkTreeRepositoryUnsupported(d)
     return False
 
 
-def find_git_dir(d):
+def find_worktree_git_dir(dotgit: PathLike) -> Optional[str]:
+    """Search for a gitdir for this worktree."""
+    try:
+        statbuf = os.stat(dotgit)
+    except OSError:
+        return None
+    if not stat.S_ISREG(statbuf.st_mode):
+        return None
+
+    try:
+        lines = open(dotgit, 'r').readlines()
+        for key, value in [line.strip().split(': ') for line in lines]:
+            if key == 'gitdir':
+                return value
+    except ValueError:
+        pass
+    return None
+
+
+def find_submodule_git_dir(d: PathLike) -> Optional[PathLike]:
+    """Search for a submodule repo."""
     if is_git_dir(d):
         return d
 
     try:
         with open(d) as fp:
             content = fp.read().rstrip()
-    except (IOError, OSError):
+    except IOError:
         # it's probably not a file
         pass
     else:
         if content.startswith('gitdir: '):
             path = content[8:]
-            if not os.path.isabs(path):
-                path = join(dirname(d), path)
-            return find_git_dir(path)
+
+            if Git.is_cygwin():
+                ## Cygwin creates submodules prefixed with `/cygdrive/...` suffixes.
+                path = decygpath(path)
+            if not osp.isabs(path):
+                path = osp.normpath(osp.join(osp.dirname(d), path))
+            return find_submodule_git_dir(path)
     # end handle exception
     return None
 
 
-def short_to_long(odb, hexsha):
+def short_to_long(odb: 'GitCmdObjectDB', hexsha: AnyStr) -> Optional[bytes]:
     """:return: long hexadecimal sha1 from the given less-than-40 byte hexsha
         or None if no candidate could be found.
     :param hexsha: hexsha with less than 40 byte"""
@@ -73,14 +114,15 @@ def short_to_long(odb, hexsha):
     # END exception handling
 
 
-def name_to_object(repo, name, return_ref=False):
+def name_to_object(repo: 'Repo', name: str, return_ref: bool = False
+                   ) -> Union[SymbolicReference, 'Commit', 'TagObject', 'Blob', 'Tree']:
     """
     :return: object specified by the given name, hexshas ( short and long )
         as well as references are supported
     :param return_ref: if name specifies a reference, we will return the reference
         instead of the object. Otherwise it will raise BadObject or BadName
     """
-    hexsha = None
+    hexsha = None  # type: Union[None, str, bytes]
 
     # is it a hexsha ? Try the most common ones, which is 7 to 40
     if repo.re_hexsha_shortened.match(name):
@@ -120,8 +162,8 @@ def name_to_object(repo, name, return_ref=False):
     return Object.new_from_sha(repo, hex_to_bin(hexsha))
 
 
-def deref_tag(tag):
-    """Recursively dereerence a tag and return the resulting object"""
+def deref_tag(tag: Tag) -> 'TagObject':
+    """Recursively dereference a tag and return the resulting object"""
     while True:
         try:
             tag = tag.object
@@ -131,7 +173,7 @@ def deref_tag(tag):
     return tag
 
 
-def to_commit(obj):
+def to_commit(obj: Object) -> Union['Commit', 'TagObject']:
     """Convert the given object to a commit if possible and return it"""
     if obj.type == 'tag':
         obj = deref_tag(obj)
@@ -142,7 +184,7 @@ def to_commit(obj):
     return obj
 
 
-def rev_parse(repo, rev):
+def rev_parse(repo: 'Repo', rev: str) -> Union['Commit', 'Tag', 'Tree', 'Blob']:
     """
     :return: Object at the given revision, either Commit, Tag, Tree or Blob
     :param rev: git-rev-parse compatible revision specification as string, please see
@@ -158,7 +200,7 @@ def rev_parse(repo, rev):
         raise NotImplementedError("commit by message search ( regex )")
     # END handle search
 
-    obj = None
+    obj = cast(Object, None)   # not ideal. Should use guards
     ref = None
     output_type = "commit"
     start = 0
@@ -208,7 +250,7 @@ def rev_parse(repo, rev):
                     pass    # error raised later
                 # END exception handling
             elif output_type in ('', 'blob'):
-                if obj.type == 'tag':
+                if obj and obj.type == 'tag':
                     obj = deref_tag(obj)
                 else:
                     # cannot do anything for non-tags
@@ -221,16 +263,16 @@ def rev_parse(repo, rev):
                 try:
                     # transform reversed index into the format of our revlog
                     revlog_index = -(int(output_type) + 1)
-                except ValueError:
+                except ValueError as e:
                     # TODO: Try to parse the other date options, using parse_date
                     # maybe
-                    raise NotImplementedError("Support for additional @{...} modes not implemented")
+                    raise NotImplementedError("Support for additional @{...} modes not implemented") from e
                 # END handle revlog index
 
                 try:
                     entry = ref.log_entry(revlog_index)
-                except IndexError:
-                    raise IndexError("Invalid revlog index: %i" % revlog_index)
+                except IndexError as e:
+                    raise IndexError("Invalid revlog index: %i" % revlog_index) from e
                 # END handle index out of bound
 
                 obj = Object.new_from_sha(repo, hex_to_bin(entry.newhexsha))
@@ -243,8 +285,8 @@ def rev_parse(repo, rev):
 
             # empty output types don't require any specific type, its just about dereferencing tags
             if output_type and obj.type != output_type:
-                raise ValueError("Could not accomodate requested object type %r, got %s" % (output_type, obj.type))
-            # END verify ouput type
+                raise ValueError("Could not accommodate requested object type %r, got %s" % (output_type, obj.type))
+            # END verify output type
 
             start = end + 1                   # skip brace
             parsed_to = start
@@ -273,11 +315,11 @@ def rev_parse(repo, rev):
         # END number parsing only if non-blob mode
 
         parsed_to = start
-        # handle hiererarchy walk
+        # handle hierarchy walk
         try:
             if token == "~":
                 obj = to_commit(obj)
-                for item in xrange(num):
+                for _ in range(num):
                     obj = obj.parents[0]
                 # END for each history item to walk
             elif token == "^":
@@ -294,8 +336,10 @@ def rev_parse(repo, rev):
             else:
                 raise ValueError("Invalid token: %r" % token)
             # END end handle tag
-        except (IndexError, AttributeError):
-            raise BadName("Invalid revision spec '%s' - not enough parent commits to reach '%s%i'" % (rev, token, num))
+        except (IndexError, AttributeError) as e:
+            raise BadName(
+                "Invalid revision spec '%s' - not enough "
+                "parent commits to reach '%s%i'" % (rev, token, num)) from e
         # END exception handling
     # END parse loop
 
